#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  unroll.py
#  
#  Copyright 2018 Ulrich Brammer <ulrich@Pauline>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  


''' To do:
- Use the same method for angle calculation for rotations due to cylin-
  drical faces. (use the uv-parameters) See FaceTree calculation of 
  self.secondRotAngle Done!
- Check the angle list for angle steps larger than pi and correct all values
  in case.
  

Revisions
01.10.2018 proof of concept. Can unfold convex shapes made from
flat faces, cylindrical faces and conical faces.
  
'''



import Part, math
from FreeCAD import Base
#import DraftVecUtils

myBox = Part.makeBox(100.0, 200.0, 300.0)

#Part.show(myBox, 'myBox')

myCyl = Part.makeCylinder(50.0, 225.0)

#Part.show(myCyl, 'myCylinder')


def equal_vertex(vert1, vert2, p=5):
  # compares two vertices 
  return (round(vert1.X - vert2.X,p)==0 and round(vert1.Y - vert2.Y,p)==0 and round(vert1.Z - vert2.Z,p)==0)

def equal_vector(vec1, vec2, p=5):
  # compares two vectors 
  return (round(vec1.x - vec2.x,p)==0 and round(vec1.y - vec2.y,p)==0 and round(vec1.z - vec2.z,p)==0)

def radial_vector(point, axis_pnt, axis):
  chord = axis_pnt.sub(point)
  norm = axis.cross(chord)
  perp = axis.cross(norm)
  # FreeCAD.Console.PrintLog( str(chord) + ' ' + str(norm) + ' ' + str(perp)+'\n')
  # print ( str(chord) + ' ' + str(norm) + ' ' + str(perp)+'\n')
  #test_line = Part.makeLine(axis_pnt.add(dist_rv),axis_pnt)
  # test_line = Part.makeLine(axis_pnt.add(perp),axis_pnt)
  # test_line = Part.makeLine(point, axis_pnt)
  # Part.show(test_line)
  return perp.normalize()

def cone_norm(point, apex, axis):
  chord = apex.sub(point)
  #norm = axis.cross(chord)
  rad = radial_vector(point, apex, axis)
  norm = rad.cross(chord)
  perp = chord.cross(norm)
  # FreeCAD.Console.PrintLog( str(chord) + ' ' + str(norm) + ' ' + str(perp)+'\n')
  # print ( str(chord) + ' ' + str(norm) + ' ' + str(perp)+'\n')
  test_line = Part.makeLine(point.add(perp),point)
  # test_line = Part.makeLine(point, axis_pnt)
  #Part.show(test_line, 'ConeNormal')
  return perp.normalize()


def cone_test(point, apex, axis, testVec, p=5):
  chord = apex.sub(point)
  norm = axis.cross(chord)
  #rad = radial_vector(point, apex, axis)
  #norm = rad.cross(chord)
  perp = chord.cross(norm)
  # FreeCAD.Console.PrintLog( str(chord) + ' ' + str(norm) + ' ' + str(perp)+'\n')
  # print ( str(chord) + ' ' + str(norm) + ' ' + str(perp)+'\n')
  test_line = Part.makeLine(point.add(perp),point)
  # test_line = Part.makeLine(point, axis_pnt)
  #Part.show(test_line, 'ConeTestNormal')
  result = testVec - perp.normalize()
  #print 'cone test result: ', result,
  #print 'cone NormVec ', testVec, ' test ', perp.normalize()
  return (round(result.x,p)==0 and round(result.y,p)==0 and round(result.z,p)==0)


def rotateVector(vec, phi, rAxis):
  ''' rotate a vector by the angle phi around the axis rAxis'''
  #https://de.wikipedia.org/wiki/Drehmatrix
  rVec = rAxis.cross(vec).cross(rAxis).multiply(math.cos(phi)) + rAxis.cross(vec)*math.sin(phi) + rAxis*rAxis.dot(vec) 
  return rVec


class FaceTree(object):
  ''' This class defines the nodes of a tree. The tree is constructed by
  the functions growTree or growBranches of the Unfolder class.
  Each face of the part gets a node in the tree.
  The indexes are the number of the face in the original part.
  '''
  def __init__(self, unfoldShape = None, f_idx=None, Parent_node= None, Parent_edge = None):
    ''' unfoldShape is the Shape to be unfolded.
    '''
    
    if unfoldShape is not None: # fix me a shape is needed to unfold
      theFace = unfoldShape.Faces[f_idx]
    else:
      print ('need a shape to unfold! Select one first!')
    self.idx = f_idx  # index of the face
    F_type = str(theFace.Surface)
    print ('analyzing Face'+str(f_idx+1), ' ', F_type)
    
    self.node_type = None  # 'Flat' or 'Cylinder' or 'Cone', see below
    self.p_node = Parent_node   # Parent node
    self.p_edge = Parent_edge # the connecting edge to the parent node
    self.child_list = [] # List of child-nodes = link to tree structure
    self.child_idx_list = [] # List of child_idx
    # need also a list of indices of child faces
    self.normal = None # normal of the face, used to calculate the rotation, calculated below
    self.minusTan = None # flag to correct the self.secondRotAngle
    
    if Parent_node is None:
      self.axis = None # the rotation axis between child and parent
      self.rotCenter = None # the rotation origin
      self.rot_angle = None # rotation angle in radians
    else:
      curveType = str(self.p_edge.Curve)
      if "<Line" in curveType:
        edgeParameter = self.p_edge.FirstParameter
      else:
        edgeParameter = (self.p_edge.FirstParameter + self.p_edge.LastParameter) / 2.0

      self.axis = self.p_edge.tangentAt(edgeParameter) # the rotation axis
      self.rotCenter = self.p_edge.valueAt(edgeParameter) # the rotation origin
      self.rot_angle = None # rotation angle in radians
    
    self.cAxis = None # Axis of a cylindrical or conical face
    self.cCenter = None 
    self.radius = None # radius of a cylinder or whatever needs a radius.
    self.analysis_ok = True # indicator if something went wrong with the analysis of the face
    self.error_code = None # index to unfold_error dictionary
    
    self.translation = None # needed if parent is an unbended face like cylinder
    self.secondRotAngle = None # rotation around rotation axis of parent
    self.thirdRotAngle = None # third rotation around normal axis of parent for cone childs
    self.reversCone = None # True if apex is on the "wrong" side
    
    s_Ori = theFace.Orientation


    F_type = str(theFace.Surface)
    if F_type == "<Plane object>":
      self.node_type = 'Flat' # fixme
      FreeCAD.Console.PrintLog("Face"+ str(f_idx+1) + " Type: "+ str(self.node_type) + "\n")
    
      #s_Ori = theFace.Orientation
      #s_Axis = theFace.Surface.Axis # is not reliable in all cases
      s_Axis = theFace.Surface.normal(0.0,0.0)
      if s_Ori == 'Forward':
        self.normal = Base.Vector(s_Axis.x, s_Axis.y, s_Axis.z)
      else:                    
        self.normal = Base.Vector(-s_Axis.x, -s_Axis.y, -s_Axis.z)
      print ('Orientation: ', s_Ori, ' Type: ', str(self.node_type), ' normal ', self.normal)

      #s_Posi = theFace.Surface.Position
      s_Posi = theFace.Edges[0].Vertexes[0].Point
      normal_line = Part.makeLine(s_Posi + self.normal, s_Posi)
      #Part.show(normal_line, 'normal_line'+str(f_idx+1)+'_')



    if F_type == "<Cylinder object>":
      self.node_type = 'Cylinder' 
      FreeCAD.Console.PrintLog("Face"+ str(f_idx+1) + " Type: "+ str(self.node_type) + "\n")
    
      #s_Ori = self._Shape.Faces[f_idx].Orientation
      s_Center = theFace.Surface.Center
      self.cCenter = s_Center
      self.radius = theFace.Surface.Radius
      self.cAxis = theFace.Surface.Axis #fix me need to understand the minus sign?
      if Parent_node is None:
        angle_0 = theFace.ParameterRange[0]
        length_0 = theFace.ParameterRange[2]
        #s_Posi = theFace.Surface.value(angle_0, length_0)
        #s_Posi = theFace.Edges[0].Vertexes[0]
        self.rotCenter = theFace.Edges[0].Vertexes[0].Point
        self.axis = self.cAxis

      #self.normal = radial_vector(self.rotCenter, s_Center, self.cAxis)
      #need to check the direction of the cylinder axis
      self.u0, self.v0 = theFace.Surface.parameter(self.rotCenter)
      s_Axis = theFace.Surface.normal(self.u0, self.v0)
      if s_Ori == 'Forward':
        self.normal = Base.Vector(s_Axis.x, s_Axis.y, s_Axis.z)
      else:                    
        self.normal = Base.Vector(-s_Axis.x, -s_Axis.y, -s_Axis.z)
      
      uTest = self.u0 + 0.1
      testPnt = theFace.Surface.value(uTest, self.v0)
      tripleProd = self.normal.cross(self.cAxis).dot(testPnt-self.rotCenter)
      #print 'tripleProd: ', tripleProd
      self.tanVec = self.normal.cross(self.cAxis)
      if tripleProd < 0:
        #self.cAxis = -self.cAxis
        self.tanVec = self.tanVec.multiply(-1.0) #self.normal.cross(self.cAxis)
        self.minusTan = True
        
      #axis_line = Part.makeLine(self.rotCenter + self.normal, self.rotCenter)
      #Part.show(axis_line, 'normal_line'+str(f_idx+1)+'_')

    if F_type == "<Cone object>":
      self.node_type = 'Cone' # 
      FreeCAD.Console.PrintLog("Face"+ str(f_idx+1) + " Type: "+ str(self.node_type) + "\n")
      
      #s_Ori = self._Shape.Faces[f_idx].Orientation
      s_Center = theFace.Surface.Center
      self.cCenter = s_Center
      self.apex = theFace.Surface.Apex
      self.semiAngle = theFace.Surface.SemiAngle
      #self.radius = theFace.Surface.Radius
      self.cAxis = theFace.Surface.Axis #fix me need to understand the minus sign?
      if Parent_node is None:
        #search for an edge that is not a circle, because it could be the apex
        for edge in theFace.OuterWire.Edges:
          eType = str(edge.Curve)
          if not ("Circle" in eType):
            coneEdge = edge
            break
        self.rotCenter = coneEdge.Vertexes[0].Point
        self.axis = self.cAxis
        # test if self.rotCenter is identical with self.apex
        if equal_vector(self.rotCenter, self.apex):
          #self.p_edge.FirstParameter
          self.rotCenter = coneEdge.Vertexes[1].Point
      else:
        if equal_vector(self.rotCenter, self.apex):
          #self.p_edge.FirstParameter
          self.rotCenter = self.p_edge.valueAt(self.p_edge.LastParameter)
          self.axis = self.p_edge.tangentAt(self.p_edge.LastParameter) # the rotation axis
      #self.normal = radial_vector(self.rotCenter, s_Center, self.cAxis)
      self.normal = cone_norm(self.rotCenter, self.apex, self.cAxis)
      if cone_test(self.rotCenter, self.apex, self.cAxis, self.normal):
        print ('cone test True')
        self.reversCone = True
      else:
        print ('cone test False')
        self.reversCone = False
      #need to check the direction of the cylinder axis
      self.u0, self.v0 = theFace.Surface.parameter(self.rotCenter)
      uTest = self.u0 + 0.1
      testPnt = theFace.Surface.value(uTest, self.v0)
      tripleProd = self.normal.cross(self.cAxis).dot(testPnt-self.rotCenter)
      print ('cone tripleProd: ', tripleProd)
      #self.tanVec = self.normal.cross(self.cAxis)
      if tripleProd < 0:
        #self.tanVec = -self.tanVec #self.normal.cross(self.cAxis)
        self.minusTan = True
      self.v0 = (self.cCenter-self.apex).Length / math.cos(self.semiAngle)



    
    if Parent_node is not None:
      #parent_node.child_list.append(tNode)
      #Bestimme die Aufbiegeoperation, nur gebogene Flächen
      #Bestimme Rotation und Versatz
      
      # calculate the angle of rotation to align to the parent face
      if self.p_node.node_type == 'Flat':
        pVec = self.p_node.normal # normal of the parent face
      elif self.p_node.node_type == 'Cylinder':
        #cyl_Center = unfoldShape.Faces[self.p_node.idx].Surface.Center
        # print 'parent: Face', str(self.p_node.idx +1), ' Type: ', self.p_node.node_type
        # print 'p cCenter: ', self.p_node.cCenter, ' axis: ', self.p_node.cAxis
        pVec = radial_vector(self.rotCenter, self.p_node.cCenter, self.p_node.cAxis) # fix me for cone!
      else:
        #cyl_Center = unfoldShape.Faces[self.p_node.idx].Surface.Center
        print ('parent: Face', str(self.p_node.idx +1), ' Type: ', self.p_node.node_type)
        #print 'p cCenter: ', self.p_node.cCenter, ' axis: ', self.p_node.cAxis
        pVec = cone_norm(self.rotCenter, self.p_node.apex, self.p_node.cAxis) 
      myVec = self.normal # normal of the flat face
      self.rot_angle = math.atan2(pVec.cross(myVec).dot(self.axis), pVec.dot(myVec))
      print ('the rot_angle: ', math.degrees(self.rot_angle))
      
      if self.p_node.node_type == 'Cylinder':
        # extra translation if parent of type Cylinder

        # fix me: I do not understand how the direction of the cAxis has to be determined?
        axis = self.p_node.cAxis # cylinderAxis can be parallel or angled to rotAxis.
    
        #cyl_Center = unfoldShape.Faces[self.p_node.idx].Surface.Center
        cyl_Center = self.p_node.cCenter
        bRad = self.p_node.radius # radius of the cylindrical face
        tanVec = self.p_node.tanVec
        nullVec = self.p_node.normal
        poi = self.rotCenter
        radVec = radial_vector(poi, cyl_Center, axis)
        angle = math.atan2(nullVec.cross(radVec).dot(axis), nullVec.dot(radVec))

        u, v = unfoldShape.Faces[self.p_node.idx].Surface.parameter(self.rotCenter)
        vVec = Base.Vector(axis.x,axis.y,axis.z) # make a copy
        uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
        # print "u: ",u, " v: ",v
        uAngle = u-self.p_node.u0
        # print 'Face'+str(f_idx+1),' angle ', math.degrees(angle), ' uAngle ', math.degrees(uAngle)
        v0 = self.p_node.v0
        uPosi = self.p_node.rotCenter + uVec.multiply(uAngle*bRad) + vVec.multiply(v-v0)

        # print 'translation from ', poi, ' to ', uPosi
        self.translation = uPosi - poi
        #self.secondRotAngle = angle
        if not self.p_node.minusTan:
          self.secondRotAngle = uAngle
        else:
          self.secondRotAngle = -uAngle

      elif self.p_node.node_type == 'Cone':
        # extra translation if parent of type Cone

        # fix me: I do not understand how the direction of the cAxis has to be determined?
        axis = self.p_node.cAxis # cylinderAxis can be parallel or angled to rotAxis.
    
        coneFace = unfoldShape.Faces[self.p_node.idx]
        con_Center = self.p_node.cCenter

        #tanVec = self.p_node.tanVec
        nullVec = self.p_node.normal
        poi = self.rotCenter
        radVec = radial_vector(poi, con_Center, axis)
        angle = math.atan2(nullVec.cross(radVec).dot(axis), nullVec.dot(radVec))

        u, v = coneFace.Surface.parameter(self.rotCenter)
        # print "u: ",u, " v: ",v
        uAngle = u-self.p_node.u0
        print ('Face'+str(f_idx+1),' angle ', math.degrees(angle), ' uAngle ', math.degrees(uAngle))

        vVec = (self.p_node.rotCenter - self.p_node.apex).normalize()
  
        #def unbendPnt(uP, vP):
        # distance to apex
        if self.p_node.reversCone:
          apexDist = self.p_node.v0 - v
        else:
          apexDist = self.p_node.v0 + v
        theVec = vVec * apexDist
        #calculate a rotation angle
        # the angle uP-u0 need to be transformed
        phi = (uAngle)*math.sin(self.p_node.semiAngle)
        #self.secondRotAngle = angle
        #self.thirdRotAngle = phi
        if not self.p_node.minusTan:
          self.secondRotAngle = uAngle
          self.thirdRotAngle = -phi
          rotVec = rotateVector(theVec, phi, self.p_node.normal)
        else:
          self.secondRotAngle = -uAngle
          self.thirdRotAngle = phi
          rotVec = rotateVector(theVec, -phi, self.p_node.normal)
        uPosi = self.p_node.apex + rotVec
        print ('translation from ', poi, ' to ', uPosi)
        self.translation = uPosi - poi


  def get_Face_idx(self):
    # get the face index from the tree-element
    return self.idx

class Unfolder(object):
  ''' This class builds the FaceTree with all information in its nodes
  to make an unfolded FaceTree.
  '''

  def __init__(self):
    self.root = None # getObject adds the root node if parent_node == None
    self._Shape = None

    self.error_code = None
    self.failed_face_idx = None
    self.f_list = [] # list with all faces of the object
    self.idx_list = [] # list with the face indexes
    

  def getObject(self, TheShape, f_idx):
    self._Shape = TheShape.copy()

    if not self._Shape.isValid():
      FreeCAD.Console.PrintLog("The shape is not valid!" + "\n")
      self.error_code = 4  # Starting: invalid shape
      self.failed_face_idx = f_idx

    self.f_list = self._Shape.Faces[:] 
    self.idx_list = list(range(len (self._Shape.Faces)))


  def growTree(self, f_idx = 0, parent_node = None, parent_edge = None):   #Funktion Baumzweig(ein Flächenindex)
    #Nimm die Fläche.
    #trage die Fläche in den Baum ein.
    tNode = FaceTree(self._Shape, f_idx, parent_node, parent_edge)
    # This face should be a node in the tree, and is therefore known!
    # removed from the list of all unknown faces
    self.idx_list.remove(f_idx)
    # This means, it could also not be found as neighbor face anymore.

    if parent_node is None:
      self.root = tNode
    else:
      parent_node.child_list.append(tNode)
      #Bestimme die Aufbiegeoperation, nur gebogene Flächen
      #Bestimme Rotation und Versatz
      
    print ('idx_list: ', self.idx_list)
    such_list = self.idx_list[:] 

    #Mache eine Liste aller Kanten.
    if self.error_code is None:
      edge_list = []
      print (self.f_list[f_idx].Area)
      #print 'Face', f_idx+1, ' has ', len(self.f_list[f_idx].OuterWire.Edges), 'edges.'
      for n_edge in self.f_list[f_idx].OuterWire.Edges:
        if parent_edge:
          if not parent_edge.isSame(n_edge):
            edge_list.append(n_edge)
        else:
          edge_list.append(n_edge)
      #print 'edge_list edges: ', len(edge_list)

      #für jede Kante:
      for sEdge in edge_list:
        #suche die angrenzende Fläche.
        the_index = None
        for i in self.idx_list:
          for sf_edge in self.f_list[i].Edges:
            if sf_edge.isSame(sEdge):
              the_index = i
              #print 'got edge face: Face', str(i+1)
              break
          if the_index is not None:
            #Baumzweig(angrenzende Fläche)
            self.growTree(the_index, tNode, sEdge)
            break
        #print 'got face index: ', the_index


  # this function needs the node to be already in the tree
  def growBranches(self, tNode): # fix me, need rotation axis and rotation center from parent edge
    # or better just the Vertex from the parent edge that defines rotation axis and rotation center.
    f_idx = tNode.idx
    parent_edge = tNode.p_edge
    parent_node =tNode.p_node

    if parent_node is None:
      self.root = tNode
      
    print ('idx_list: ', self.idx_list)
    FreeCAD.Console.PrintLog('idx_list: '+ str(self.idx_list) + "\n")

    #Mache eine Liste aller Kanten.
    if self.error_code is None:
      edge_list = []
      print (self.f_list[f_idx].Area)
      #print 'Face', f_idx+1, ' has ', len(self.f_list[f_idx].OuterWire.Edges), 'edges.'
      for n_edge in self.f_list[f_idx].OuterWire.Edges:
        if parent_edge:
          if not parent_edge.isSame(n_edge):
            edge_list.append(n_edge)
        else:
          edge_list.append(n_edge)
      #print 'edge_list edges: ', len(edge_list)

      #für jede Kante:
      for sEdge in edge_list:
        #suche die angrenzende Fläche.
        such_list = self.idx_list[:]
        the_index = None
        for i in such_list:
          for sf_edge in self.f_list[i].Edges:
            if sf_edge.isSame(sEdge):
              the_index = i
              #print 'got edge face: Face', str(i+1)
              break
          if the_index is not None:
            #Baumzweig(angrenzende Fläche)
            #self.growTree(the_index, tNode, sEdge)
            newNode = FaceTree(self._Shape, the_index, tNode, sEdge)
            tNode.child_list.append(newNode)
            tNode.child_idx_list.append(the_index)
            self.idx_list.remove(the_index)
            break
        #print 'got face index: ', the_index
        
    for node in tNode.child_list:
      self.growBranches(node)
    FreeCAD.Console.PrintLog('Finished face'+ str(f_idx+1) + "\n")





        
  #Funktion Aufbiegen(ein Flächenknoten)
    #platte Fläche = biege Fläche auf.
    #für alle Kinder der Fläche:
      #neue Flächen = Aufbiegen (Flächenkind)
      #rotiere und versetzte neue Flächen.
      #platte Flächen = platte Flächen + neue Flächen
    #liefere alle Flächen aus.

  def unfold_tree2(self, node):
    # This function traverses the tree and unfolds the faces 
    # beginning at the outermost nodes.
    print ("unfold_tree face", node.idx + 1, ' ', node.child_idx_list)
    FreeCAD.Console.PrintLog("unfold_tree face" + str(node.idx +1) + ' '+ str(node.child_idx_list) + "\n")
    if node.node_type == 'Flat':
      theShell = [self.f_list[node.idx].copy()] # fix me! aufbiegen!
    elif node.node_type == 'Cylinder':
      #theShell = [self.unbendFace(node.idx, node, node.normal)]
      theShell = [self.unbendFace1(node.idx, node, node.normal)]
    else: #fix me  need to through an error if this is not a cone!
      theShell = [self.unbendCone(node.idx, node)]
    nodeShell = []
    FreeCAD.Console.PrintLog("got Face into the Shell" + str(node.idx +1)  + "\n")
    for n_node in node.child_list:
      if self.error_code == None:
        shell = self.unfold_tree2(n_node)
        print ('Face', node.idx+1, ' shell faces: ', len(shell))
        theShell = theShell + shell
    if node.p_node is not None:
      for bFaces in theShell:
        bFaces.rotate(node.rotCenter, node.axis, math.degrees(-node.rot_angle))
      if node.secondRotAngle is not None:
        for bFaces in theShell:
          bFaces.rotate(node.rotCenter, node.p_node.cAxis, math.degrees(node.secondRotAngle))
      if node.thirdRotAngle is not None:
        for bFaces in theShell:
          bFaces.rotate(node.rotCenter, node.p_node.normal, math.degrees(-node.thirdRotAngle))
      if node.translation is not None:
        for bFaces in theShell:
          bFaces.translate(node.translation)
        
    FreeCAD.Console.PrintLog("ufo finish face" + str(node.idx +1) + "\n")
    return (theShell + nodeShell)

  #def unrollCylinder2(self, theCyl, theLine, thick, kFactor):
  def unbendFace1(self, fIdx, bend_node, nullVec):
    '''
    This function takes a cylindrical face and returns a flat version
    '''
    theCyl = self.f_list[fIdx]
    theCenter = theCyl.Surface.Center
    #theAxis = theCyl.Surface.Axis
    theAxis =  bend_node.cAxis
    theRadius = theCyl.Surface.Radius
    umin, umax, vmin, vmax = theCyl.ParameterRange
    
    #sAxis = theCyl.valueAt(umin, vmin+1.0) - theCyl.valueAt(umin, vmin)
    #geoAxis = theCyl.valueAt(umin, vmax) - theCyl.valueAt(umin, vmin)
    #print 'sAxis: ', sAxis, ' geoAxis: ', geoAxis
  
    
    # show the vectors as lines
    
    #axisEdge = Part.makeLine(theCenter, theCenter+theAxis+theAxis+theAxis)
    #Part.show(axisEdge)
  
    oPt = bend_node.rotCenter
  
    # the normal vector of the projection plane
    nullVec = radial_vector(oPt, theCenter, theAxis)
    #radiusEdge = Part.makeLine(theCenter, theCenter+nullVec+nullVec+nullVec)
    #Part.show(radiusEdge)
    
    u0, v0 = theCyl.Surface.parameter(oPt)
    
    
    #tanVec = nullVec.cross(theAxis)
    tanVec = bend_node.tanVec
    
    wireList = []
    divisions = 12 # fix me! need a dependence on something useful.
    j=0
    for fEdge in theCyl.Edges:
      eType = str(fEdge.Curve)
      print ("the type of curve: ", eType)
      urollPts = []
      vList = []
      angles = []
      
      if ("<Ellipse object>" in eType):
        minPar, maxPar = fEdge.ParameterRange
        print ("the Parameterrange: ", minPar, " - ", maxPar, " Type: ",eType)
        iMulti = (maxPar-minPar)/divisions
        
        for i in list(range(divisions+1)):
          #print "j, i: ", j," ",i," ",fEdge.valueAt(minPar + i*iMulti)
          posi = fEdge.valueAt(minPar + i*iMulti)
          #print 'i posi: ', i, ' ', posi
          u, v = theCyl.Surface.parameter(posi)
          print ("j, i: ", j," ",i," ",u, " ", v)

          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          # print "u: ",u, " v: ",v
          angle = u-u0
          uPosi = oPt + uVec.multiply(angle*theRadius) + vVec.multiply(v-v0)
          #print "u: ",u, " v: ",v, " uPosi", uPosi, " tanVec: ", theAxis
          urollPts.append(uPosi)
        
          angles.append(angle)
          vList.append(v)
          #print 'non correction Angle: ', angle

        diffAngle0 = (angles[0] - angles[1])
        if (diffAngle0 > math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[0]= oPt + uVec.multiply((angles[0]-2.0*math.pi)*theRadius) + vVec.multiply(vList[0]-v0)
          print ('old angle: ', angles[0], ' corrected angle: ', angles[0]-2.0*math.pi)
        if (diffAngle0 < -math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[0]= oPt + uVec.multiply((angles[0]+2.0*math.pi)*theRadius) + vVec.multiply(vList[0]-v0)
          print ('old angle: ', angles[0], ' corrected angle: ', angles[0]+2.0*math.pi)

        diffAngle1 = (angles[divisions] - angles[divisions-1])
        if (diffAngle1 > math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[divisions]= oPt + uVec.multiply((angles[divisions]-2.0*math.pi)*theRadius) + vVec.multiply(vList[divisions]-v0)
          print ('old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]-2.0*math.pi)
        if (diffAngle1 < -math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[divisions]= oPt + uVec.multiply((angles[divisions]+2.0*math.pi)*theRadius) + vVec.multiply(vList[divisions]-v0)
          print ('old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]+2.0*math.pi)
        #myPoly = Part.makePolygon ( urollPts )
        #Part.show(myPoly, 'ElliPoly')
        
        uCurve = Part.BSplineCurve()
        uCurve.interpolate(urollPts)
        theCurve = uCurve.toShape()
        wireList.append(theCurve)
        #Part.show(theCurve, 'Elli')
        
      if "<Line" in eType:
        
        if fEdge.isSeam(theCyl):
          
          uL0, vL0 = theCyl.Surface.parameter(fEdge.Vertexes[0].Point)
          uL1, vL1 = theCyl.Surface.parameter(fEdge.Vertexes[1].Point)
          pnt0 = oPt + tanVec*((umin-u0)*theRadius) + theAxis*(vL0-v0)
          pnt1 = oPt + tanVec*((umin-u0)*theRadius) + theAxis*(vL1-v0)
          edgeL = Part.makeLine(pnt0, pnt1)
          print ('pnt0 ', pnt0, ' pnt1 ', pnt1)
          pnt0 = oPt + tanVec*((umax-u0)*theRadius) + theAxis*(vL0-v0)
          pnt1 = oPt + tanVec*((umax-u0)*theRadius) + theAxis*(vL1-v0)
          edgeSeam = Part.makeLine(pnt0, pnt1)
          print ('Seam pnt0 ', pnt0, ' pnt1 ', pnt1)
          #lWire = Part.Wire([edgeL])
          wireList.append(edgeL)
          wireList.append(edgeSeam)
          #Part.show(edgeSeam, 'Seamline')
        else: # need to distinguish betweeen angle 0 and 2*pi
          uL0, vL0 = theCyl.Surface.parameter(fEdge.Vertexes[0].Point)

          if (((uL0 + 0.001) > 0.0) and ((uL0 - 0.001) < 0.0)) or \
            (((uL0 + 0.001) > 2.0*math.pi) and ((uL0 - 0.001) < 2.0*math.pi)): 
            print ('got condition')
            # fix me, may fail if range is not between 0.0 and 2*math.pi
            uL1, vL1 = theCyl.Surface.parameter(fEdge.Vertexes[1].Point)
            pnt = theCyl.Surface.value(0.01, (vL0 + vL1)/2.0)
            if theCyl.isInside(pnt,0.01,True):
              u = 0.0
            else:
              u = 2*math.pi
            pnt0 = oPt + tanVec*((u-u0)*theRadius) + theAxis*(vL0-v0)
            pnt1 = oPt + tanVec*((u-u0)*theRadius) + theAxis*(vL1-v0)
            edgeL = Part.makeLine(pnt0, pnt1)
            
          else:
            urollPts = []
            for lVert in fEdge.Vertexes:
              
              u, v = theCyl.Surface.parameter(Base.Vector(lVert.X, lVert.Y, lVert.Z))
              #print "u, v: ",u, " ", v
              vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
              uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
              uPosi = oPt + uVec.multiply((u-u0)*theRadius) + vVec.multiply(v-v0)
              # print "u: ",u, " v: ",v, " uPosi", uPosi, " tanVec: ", theAxis
              urollPts.append(uPosi)
            edgeL = Part.makeLine(urollPts[0], urollPts[1])
          wireList.append(edgeL)
          #Part.show(lWire, 'Line')
        
        
      if "Circle" in eType:
        parList = fEdge.ParameterRange
        print ("the Parameterrange: ", parList[0], " , ", parList[1], " Type: ",eType)
        posi = fEdge.valueAt(parList[0])
        u, v = theCyl.Surface.parameter(posi)
        vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
        uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
        pnt0 = oPt + uVec*((parList[0]-u0)*theRadius) + vVec*(v-v0)
        pnt1 = oPt + uVec*((parList[1]-u0)*theRadius) + vVec*(v-v0)
        edgeL = Part.makeLine(pnt0, pnt1)
        '''
        for para in parList:
          #print "parameter: ",fEdge.valueAt(para)
          posi = fEdge.valueAt(para)
          u, v = theCyl.Surface.parameter(posi)
          # print "u, v: ",u, " ", v
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          # print "u: ",u, " v: ",v
          uPosi = oPt + uVec.multiply((u-u0)*theRadius) + vVec.multiply(v-v0)
          print "u: ",u, " v: ",v, " uPosi", uPosi, " tanVec: ", theAxis
          urollPts.append(uPosi)
        edgeL = Part.makeLine(urollPts[0], urollPts[1])
        '''
        wireList.append(edgeL)
  
      if ("<BSplineCurve object>" in eType):
        minPar, maxPar = fEdge.ParameterRange
        print ("the Parameterrange: ", minPar, " - ", maxPar, " Type: ",eType)
        iMulti = (maxPar-minPar)/divisions
        for i in list(range(divisions+1)):
          posi = fEdge.valueAt(minPar + i*iMulti)
          #print 'i posi: ', i, ' ', posi
          u, v = theCyl.Surface.parameter(posi)
          #print "j, i: ", j," ",i," ",u, " ", v
          #print "i: ", i," ",minPar + i*iMulti, ' u-u0: ', u-u0
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          # print "u: ",u, " v: ",v
          angle = u-u0
          uPosi = oPt + uVec.multiply(angle*theRadius) + vVec.multiply(v-v0)
          #print "u: ",u, " v: ",v, " uPosi", uPosi, " tanVec: ", theAxis
          urollPts.append(uPosi)

          angles.append(angle)
          vList.append(v)
          #print 'non correction Angle: ', angle
        diffAngle0 = (angles[0] - angles[1])
        if (diffAngle0 > math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[0]= oPt + uVec.multiply((angles[0]-2.0*math.pi)*theRadius) + vVec.multiply(vList[0]-v0)
          #print 'old angle: ', angles[0], ' corrected angle: ', angles[0]-2.0*math.pi
        if (diffAngle0 < -math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[0]= oPt + uVec.multiply((angles[0]+2.0*math.pi)*theRadius) + vVec.multiply(vList[0]-v0)
          #print 'old angle: ', angles[0], ' corrected angle: ', angles[0]+2.0*math.pi

        diffAngle1 = (angles[divisions] - angles[divisions-1])
        if (diffAngle1 > math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[divisions]= oPt + uVec.multiply((angles[divisions]-2.0*math.pi)*theRadius) + vVec.multiply(vList[divisions]-v0)
          #print 'old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]-2.0*math.pi
        if (diffAngle1 < -math.pi):
          vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          urollPts[divisions]= oPt + uVec.multiply((angles[divisions]+2.0*math.pi)*theRadius) + vVec.multiply(vList[divisions]-v0)
          print ('old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]+2.0*math.pi)


        
        myPoly = Part.makePolygon ( urollPts )
        #Part.show(myPoly, 'B_spline_poly')
        
        uCurve = Part.BSplineCurve()
        uCurve.interpolate(urollPts)
        theCurve = uCurve.toShape()
        wireList.append(theCurve)
        #Part.show(theCurve, 'B_spline')
  
        
      j += 1
    eList = Part.__sortEdges__(wireList)
    myWire = Part.Wire(eList)
    myCompound = Part.Compound(wireList)
    Part.show(myCompound, 'CylinderCompound')
    theFace = Part.Face(myWire)
    #Part.show(theFace)
    return theFace


  def unbendCone(self, fIdx, bend_node):
    '''
    This function takes a cylindrical face and returns a flat version
    '''
    theCone = self.f_list[fIdx]
    theCenter = bend_node.apex
    #theAxis = theCyl.Surface.Axis
    theAxis =  bend_node.cAxis
    #theRadius = theCyl.Surface.Radius
    umin, umax, vmin, vmax = theCone.ParameterRange
    
    # show the vectors as lines
    
    #axisEdge = Part.makeLine(theCenter, theCenter+theAxis+theAxis+theAxis)
    #Part.show(axisEdge)
  
    u0 = bend_node.u0
    print ('Cone u0 ', u0)
    #get the distance to the apex at v = 0.0
    v0 = bend_node.v0
    vVec = (bend_node.rotCenter - bend_node.apex).normalize()
    
    def unbendPnt(uP, vP):
      # distance to apex
      #apexDist = v0+vP
      if not bend_node.reversCone:
        apexDist = v0+vP
      else:
        apexDist = v0-vP

      #calculate a rotation angle
      # the angle uP-u0 need to be transformed
      if bend_node.minusTan:
        phi = -(uP-u0)*math.sin(bend_node.semiAngle)
      else:
        phi = (uP-u0)*math.sin(bend_node.semiAngle)
        
      theVec = vVec * apexDist
      
      rotVec = self.rotateVec(theVec, phi, bend_node.normal)
      return bend_node.apex + rotVec
    
    
    
    wireList = []
    divisions = 12 # fix me! need a dependence on something useful.
    j=0
    for fEdge in theCone.Edges:
      eType = str(fEdge.Curve)
      print ("the type of curve: ", eType)
      urollPts = []
      vList = []
      angles = []
      
      if ("<Ellipse object>" in eType) or ("<Hyperbola object>" in eType) or ("<Parabola object>" in eType):
        minPar, maxPar = fEdge.ParameterRange
        #print "the Parameterrange: ", minPar, " - ", maxPar, " Type: ",eType
        iMulti = (maxPar-minPar)/divisions
        
        for i in list(range(divisions+1)):
          #print "j, i: ", j," ",i," ",fEdge.valueAt(minPar + i*iMulti)
          posi = fEdge.valueAt(minPar + i*iMulti)
          #print 'i posi: ', i, ' ', posi
          u, v = theCone.Surface.parameter(posi)
          #print "j, i: ", j," ",i," ",u, " ", v
          ## print "u: ",u, " v: ",v
          angle = u-u0
          uPosi = unbendPnt(u, v)
          #print "u: ",u, " v: ",v, " uPosi", uPosi, " tanVec: ", theAxis
          urollPts.append(uPosi)
        
          angles.append(angle)
          vList.append(v)
          #print 'non correction Angle: ', angle

        diffAngle0 = (angles[0] - angles[1])
        if (diffAngle0 > math.pi):
          urollPts[0] = unbendPnt(angles[0]-2.0*math.pi + u0, vList[0])
          print ('old angle: ', angles[0], ' corrected angle: ', angles[0]-2.0*math.pi)
        if (diffAngle0 < -math.pi):
          urollPts[0] = unbendPnt(angles[0]+2.0*math.pi + u0, vList[0])
          print ('old angle: ', angles[0], ' corrected angle: ', angles[0]+2.0*math.pi)

        diffAngle1 = (angles[divisions] - angles[divisions-1])
        if (diffAngle1 > math.pi):
          urollPts[0] = unbendPnt(angles[divisions]-2.0*math.pi + u0, vList[divisions])
          print ('old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]-2.0*math.pi)
        if (diffAngle1 < -math.pi):
          urollPts[0] = unbendPnt(angles[divisions]+2.0*math.pi + u0, vList[divisions])
          print ('old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]+2.0*math.pi)
        #myPoly = Part.makePolygon ( urollPts )
        #Part.show(myPoly, 'ElliPoly')
        
        uCurve = Part.BSplineCurve()
        uCurve.interpolate(urollPts)
        theCurve = uCurve.toShape()
        wireList.append(theCurve)
        #Part.show(theCurve, 'Elli')
        
      if "<Line" in eType:
        
        if fEdge.isSeam(theCone):
          
          uL0, vL0 = theCone.Surface.parameter(fEdge.Vertexes[0].Point)
          uL1, vL1 = theCone.Surface.parameter(fEdge.Vertexes[1].Point)
          pnt0 = unbendPnt(umin, vL0)
          pnt1 = unbendPnt(umin, vL1)
          edgeL = Part.makeLine(pnt0, pnt1)
          #print 'pnt0 ', pnt0, ' pnt1 ', pnt1
          pnt0 = unbendPnt(umax, vL0)
          pnt1 = unbendPnt(umax, vL1)
          edgeSeam = Part.makeLine(pnt0, pnt1)
          #print 'Seam pnt0 ', pnt0, ' pnt1 ', pnt1
          wireList.append(edgeL)
          wireList.append(edgeSeam)
          #Part.show(edgeSeam, 'SeamLine')
        else: # need to distinguish betweeen angle 0 and 2*pi
          uL0, vL0 = theCone.Surface.parameter(fEdge.Vertexes[0].Point)
          uL1, vL1 = theCone.Surface.parameter(fEdge.Vertexes[1].Point)
          if equal_vector(fEdge.Vertexes[0].Point, bend_node.apex):
            uT = uL1
          else:
            uT = uL0

          if (((uT + 0.001) > 0.0) and ((uT - 0.001) < 0.0)) or \
            (((uT + 0.001) > 2.0*math.pi) and ((uT - 0.001) < 2.0*math.pi)): 
            print ('got condition')
            # fix me, may fail if range is not between 0.0 and 2*math.pi
            
            pnt = theCone.Surface.value(0.01, (vL0 + vL1)/2.0)
            if theCone.isInside(pnt,0.01,True):
              u = 0.0
            else:
              u = 2*math.pi
            pnt0 = unbendPnt(u, vL0)
            pnt1 = unbendPnt(u, vL1)
            edgeL = Part.makeLine(pnt0, pnt1)
            
          else:
            urollPts = []
            for lVert in fEdge.Vertexes:
              
              u, v = theCone.Surface.parameter(Base.Vector(lVert.X, lVert.Y, lVert.Z))
              #print "u, v: ",u, " ", v
              uPosi = unbendPnt(u, v)
              # print "u: ",u, " v: ",v, " uPosi", uPosi, " tanVec: ", theAxis
              urollPts.append(uPosi)
            edgeL = Part.makeLine(urollPts[0], urollPts[1])
          wireList.append(edgeL)
          #Part.show(edgeL, 'Line')
        
        
      if "Circle" in eType:
        parList = fEdge.ParameterRange
        #print "the Parameterrange: ", parList[0], " , ", parList[1], " Type: ",eType
        
        pos0 = fEdge.valueAt(parList[0])
        uL0, vL0 = theCone.Surface.parameter(pos0)
        pos1 = fEdge.valueAt(parList[1])
        uL1, vL1 = theCone.Surface.parameter(pos1)
        
        if not equal_vector(pos0, theCenter):
          #pnt0 = unbendPnt(uL0, vL0)
          #pnt1 = unbendPnt(uL1, vL1)
          #pntC = unbendPnt((uL0+uL1)/2.0, (vL0+vL1)/2.0)
          pnt0 = unbendPnt(parList[0], vL0)
          pnt1 = unbendPnt(parList[1], vL1)
          pntC = unbendPnt((parList[0]+parList[1])/2.0, (vL0+vL1)/2.0)
          print ('pnt0 ', pnt0, ' pnt1 ', pnt1, ' pntC ', pntC)
          #circlePoly = Part.makePolygon ( [pnt0, pntC, pnt1] )
          #Part.show(circlePoly, 'circlePoly')
          edgeL = Part.Arc(pnt0, pntC, pnt1).toShape()
          wireList.append(edgeL)
          #Part.show(edgeL, 'ConeCircle')
  
      if ("<BSplineCurve object>" in eType):
        minPar, maxPar = fEdge.ParameterRange
        print ("the Parameterrange: ", minPar, " - ", maxPar, " Type: ",eType)
        iMulti = (maxPar-minPar)/divisions
        for i in list(range(divisions+1)):
          posi = fEdge.valueAt(minPar + i*iMulti)
          #print 'i posi: ', i, ' ', posi
          u, v = theCone.Surface.parameter(posi)
          #print "j, i: ", j," ",i," ",u, " ", v
          #print "i: ", i," ",minPar + i*iMulti, ' u-u0: ', u-u0
          #vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          #uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          # print "u: ",u, " v: ",v
          angle = u-u0
          #uPosi = oPt + uVec.multiply(angle*theRadius) + vVec.multiply(v-v0)
          uPosi = unbendPnt(u, v)
          #print "u: ",u, " v: ",v, " uPosi", uPosi, " tanVec: ", theAxis
          urollPts.append(uPosi)

          angles.append(angle)
          vList.append(v)
          #print 'non correction Angle: ', angle
        diffAngle0 = (angles[0] - angles[1])
        if (diffAngle0 > math.pi):
          #vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          #uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          #urollPts[0]= oPt + uVec.multiply((angles[0]-2.0*math.pi)*theRadius) + vVec.multiply(vList[0]-v0)
          urollPts[0] = unbendPnt(angles[0]-2.0*math.pi + u0, vList[0])
          #print 'old angle: ', angles[0], ' corrected angle: ', angles[0]-2.0*math.pi
        if (diffAngle0 < -math.pi):
          #vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          #uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          #urollPts[0]= oPt + uVec.multiply((angles[0]+2.0*math.pi)*theRadius) + vVec.multiply(vList[0]-v0)
          urollPts[0] = unbendPnt(angles[0]+2.0*math.pi + u0, vList[0])
          #print 'old angle: ', angles[0], ' corrected angle: ', angles[0]+2.0*math.pi

        diffAngle1 = (angles[divisions] - angles[divisions-1])
        if (diffAngle1 > math.pi):
          #vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          #uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          #urollPts[divisions]= oPt + uVec.multiply((angles[divisions]-2.0*math.pi)*theRadius) + vVec.multiply(vList[divisions]-v0)
          urollPts[0] = unbendPnt(angles[divisions]-2.0*math.pi + u0, vList[divisions])
          #print 'old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]-2.0*math.pi
        if (diffAngle1 < -math.pi):
          #vVec = Base.Vector(theAxis.x,theAxis.y,theAxis.z) # make a copy
          #uVec = Base.Vector(tanVec.x, tanVec.y, tanVec.z)  # make a copy
          #urollPts[divisions]= oPt + uVec.multiply((angles[divisions]+2.0*math.pi)*theRadius) + vVec.multiply(vList[divisions]-v0)
          urollPts[0] = unbendPnt(angles[divisions]+2.0*math.pi + u0, vList[divisions])
          print ('old div angle: ', angles[divisions], ' corrected angle: ', angles[divisions]+2.0*math.pi)


        
        myPoly = Part.makePolygon ( urollPts )
        #Part.show(myPoly, 'B_spline_poly')
        
        uCurve = Part.BSplineCurve()
        uCurve.interpolate(urollPts)
        theCurve = uCurve.toShape()
        wireList.append(theCurve)
        #Part.show(theCurve, 'B_spline')
  
        
      j += 1
    eList = Part.__sortEdges__(wireList)
    myWire = Part.Wire(eList)
    myCompound = Part.Compound(wireList)
    Part.show(myCompound, 'EdgeCompound')
    theFace = Part.Face(myWire)
    #Part.show(theFace, 'theConeFace')
    return theFace




  def rotateVec(self, vec, phi, rAxis):
    ''' rotate a vector by the angle phi around the axis rAxis'''
    #https://de.wikipedia.org/wiki/Drehmatrix
    rVec = rAxis.cross(vec).cross(rAxis).multiply(math.cos(phi)) + rAxis.cross(vec)*math.sin(phi) + rAxis*rAxis.dot(vec) 
    return rVec



theFace = 0
s = Gui.Selection.getSelection()

#theUnfold = Unfolder()
##theUnfold.getObject(myBox, 0)
#theUnfold.getObject(s[0].Shape, theFace)
#theUnfold.growTree(theFace, None, None)
#theFaceList = theUnfold.unfold_tree2(theUnfold.root)
#try:
  #newShell = Part.Shell(theFaceList)
  #Part.show(newShell, 'theUnfold')
#except:
  #Part.show(Part.Compound(theFaceList), 'theFaces')


otherUnfold = Unfolder()
otherUnfold.getObject(s[0].Shape, theFace)
#otherUnfold.getObject(myBox, 0)
#otherUnfold.getObject(myCyl, 0)
otherUnfold.root = FaceTree(otherUnfold._Shape, theFace, None, None)
otherUnfold.idx_list.remove(theFace)
otherUnfold.growBranches(otherUnfold.root)
FreeCAD.Console.PrintLog('Analyse beendet' + "\n")

otherFaceList = otherUnfold.unfold_tree2(otherUnfold.root)
FreeCAD.Console.PrintLog('Unfold beendet' + "\n")
try:
  otherShell = Part.Shell(otherFaceList)
  Part.show(otherShell, 'OtherShell')
except:
  Part.show(Part.Compound(otherFaceList), 'OtherFaces')

if otherUnfold.error_code is not None:
  print ('Error Code: ', otherUnfold.error_code, ' Face', str(otherUnfold.failed_face_idx +1))
